/* -*- mode: c  -*- */

#include <mkcl/mkcl.h>
#include <ctype.h>
#include <limits.h>
#include <mkcl/internal.h>

#define CL_PACKAGE 0 /* 0 << 2 */
#define SI_PACKAGE 4 /* 1 << 2 */
#define EXT_PACKAGE SI_PACKAGE
#define GRAY_PACKAGE 32 /* 8 << 2 */
#define KEYWORD_PACKAGE 8 /* 2 << 2 */
#define MT_PACKAGE 12 /* 3 << 2 */
#define CLOS_PACKAGE 16 /* 4 << 2 */

#define ORDINARY_SYMBOL 0
#define CONSTANT_SYMBOL 1
#define SPECIAL_SYMBOL 2
#define FORM_SYMBOL 3

#define CL_ORDINARY	CL_PACKAGE | ORDINARY_SYMBOL
#define CL_SPECIAL	CL_PACKAGE | SPECIAL_SYMBOL
#define CL_CONSTANT	CL_PACKAGE | CONSTANT_SYMBOL
#define CL_FORM		CL_PACKAGE | ORDINARY_SYMBOL | FORM_SYMBOL
#define SI_ORDINARY	SI_PACKAGE | ORDINARY_SYMBOL
#define SI_SPECIAL	SI_PACKAGE | SPECIAL_SYMBOL
#define SI_CONSTANT	SI_PACKAGE | CONSTANT_SYMBOL
#define EXT_ORDINARY	EXT_PACKAGE | ORDINARY_SYMBOL
#define EXT_SPECIAL	EXT_PACKAGE | SPECIAL_SYMBOL
#define EXT_CONSTANT	EXT_PACKAGE | CONSTANT_SYMBOL
#define EXT_FORM	EXT_PACKAGE | ORDINARY_SYMBOL | FORM_SYMBOL
#define MT_ORDINARY	MT_PACKAGE | ORDINARY_SYMBOL
#define MT_SPECIAL	MT_PACKAGE | SPECIAL_SYMBOL
#define MT_CONSTANT	MT_PACKAGE | CONSTANT_SYMBOL
#define CLOS_ORDINARY	CLOS_PACKAGE | ORDINARY_SYMBOL
#define CLOS_SPECIAL	CLOS_PACKAGE | SPECIAL_SYMBOL
#define KEYWORD		KEYWORD_PACKAGE | CONSTANT_SYMBOL
#define GRAY_ORDINARY	GRAY_PACKAGE | ORDINARY_SYMBOL

#include "symbols_list.h"

mkcl_index mkcl_root_symbols_count = 0;

static unsigned char *
mangle_name(mkcl_object output, unsigned char *source, int l)
{
  unsigned char c;

  while (l--) {
    c = *(source++);
    if (mkcl_alphanumericp(c)) {
      c = mkcl_char_downcase(c);
    } else if (c == '-' || c == '_') {
      c = '_';
    } else if (c == '&') {
      c = 'A';
    } else if (c == '*') {
      c = 'X';
    } else if (c == '+') {
      c = 'P';
    } else if (c == '<') {
      c = 'L';
    } else if (c == '>') {
      c = 'G';
    } else if (c == '=') {
      c = 'E';
    } else if (c == '/') {
      c = 'N';
    } else if (c == ':') {
      c = 'X';
    } else {
      return NULL;
    }
    output->base_string.self[output->base_string.fillp++] = c;
  }
  return &output->base_string.self[output->base_string.fillp];
}

#define NB_STATIC_SYMBOLS (sizeof(mkcl_root_symbols)/sizeof(mkcl_root_symbols[0]))

/* Universal prefix for ManKai Common Lisp. JCB */
#define PREFIX "mk_"
#define PREFIX_SIZE (sizeof(PREFIX) - 1)

@(defun si::mangle-name (symbol &optional as_function)
  mkcl_index l;
  unsigned char c, *source, *dest;
  mkcl_object output;
  mkcl_object package;
  mkcl_object found = mk_cl_Cnil;
  mkcl_object maxarg = MKCL_MAKE_FIXNUM(MKCL_CALL_ARGUMENTS_LIMIT);
  mkcl_object minarg = MKCL_MAKE_FIXNUM(0);
  bool is_symbol;
  mkcl_object name;
@
  name = mkcl_symbol_name(env, symbol);
  is_symbol = mkcl_Null(as_function);
  if (is_symbol) {
    mkcl_word p;
    if (symbol == mk_cl_Cnil)
      { @(return mk_cl_Ct mkcl_make_simple_base_string(env, "mk_cl_Cnil")); }
    else if (symbol == mk_cl_Ct)
      { @(return mk_cl_Ct mkcl_make_simple_base_string(env, "mk_cl_Ct")); }
    p  = (mkcl_symbol_initializer*)symbol - mkcl_root_symbols;
    if (p >= 0 && p <= mkcl_root_symbols_count) {
      found = mk_cl_Ct;
      output = mk_cl_format(env, 4, mk_cl_Cnil,
			 mkcl_make_simple_base_string(env, "MKCL_SYM(~S,~D)"),
			 name, MKCL_MAKE_FIXNUM(p));
      @(return found output maxarg);
    }
  } else if (!mkcl_Null(symbol)) {
    mkcl_object fun = symbol->symbol.gfdef;
    mkcl_type t = (fun == MKCL_OBJNULL)? mkcl_t_end : mkcl_type_of(fun);
    if ((t == mkcl_t_cfun) && fun->cfun.block == MKCL_OBJNULL)
      {
	mkcl_object s = fun->cfun.name;

	if (!mkcl_Null(s)
	    && ((mkcl_object) &(mkcl_root_symbols[0].data)) <= s
	    && s < ((mkcl_object) &(mkcl_root_symbols[NB_STATIC_SYMBOLS].data)))
	  {
	    symbol = s;
	    found = mk_cl_Ct;
	    if (fun->cfun.narg >= 0)
	      {
		minarg = maxarg = MKCL_MAKE_FIXNUM(fun->cfun.narg);
	      }
	  }
      }
  }
  package = mkcl_symbol_package(env, symbol);
  if (mkcl_Null(package))
    ;
  else if (package == mkcl_core.lisp_package)
    package = mkcl_make_simple_base_string(env, "cl");
  else if (package == mkcl_core.system_package)
    package = mkcl_make_simple_base_string(env, "si");
  else if (package == mkcl_core.keyword_package)
    package = mk_cl_Cnil;
  else
    package = package->pack.name;
  symbol = mkcl_symbol_name(env, symbol);
  l = symbol->base_string.fillp;
  source = symbol->base_string.self;
  output = mkcl_alloc_simple_base_string(env, mkcl_length(env, package) + l + 1 + PREFIX_SIZE);

  if (is_symbol && source[0] == '*') {
    if (l > 2 && source[l-1] == '*') l--;
    c = 'V';
    l--;
    source++;
  } else if (is_symbol && l > 2 && source[0] == '+' && source[l-1] == '+') {
    c = 'C';
    l-= 2;
    source++;
  } else if (!is_symbol) {
    c = '_';
  } else if (package == mkcl_core.keyword_package) {
    c = 'K';
  } else {
    c = 'S';
  }
  output->base_string.fillp = 0;
  if (!mkcl_Null(package))
    {
      mangle_name(output, (unsigned char *) PREFIX, PREFIX_SIZE);
      if (!mangle_name(output, package->base_string.self, package->base_string.fillp))
	{ @(return mk_cl_Cnil mk_cl_Cnil maxarg); }
    }
  output->base_string.self[output->base_string.fillp++] = c;
  if (!(dest = mangle_name(output, source, l)))
    { @(return mk_cl_Cnil mk_cl_Cnil maxarg) }
  if (dest[-1] == '_')
    dest[-1] = 'M';
  *(dest++) = '\0';
  @(return found output minarg maxarg);
@)

static void
make_this_symbol(MKCL, int i, mkcl_object s, int code, const char *name,
		 mkcl_objectfn fun, int narg, mkcl_object value)
{
  enum mkcl_stype stp;
  mkcl_object package;
  bool form = 0;

  switch (code & 3) {
  case ORDINARY_SYMBOL: stp = mkcl_stp_ordinary; break;
  case SPECIAL_SYMBOL: stp = mkcl_stp_special; break;
  case CONSTANT_SYMBOL: stp = mkcl_stp_constant; break;
  case FORM_SYMBOL: form = 1; stp = mkcl_stp_ordinary; break;
  default: mkcl_lose("Unknown symbol type in make_this_symbol");
  }
  switch (code & ~(int)3) {
  case CL_PACKAGE: package = mkcl_core.lisp_package; break;
  case SI_PACKAGE: package = mkcl_core.system_package; break;
  case KEYWORD_PACKAGE: package = mkcl_core.keyword_package; break;
  case MT_PACKAGE: package = mkcl_core.mt_package; break;
  case CLOS_PACKAGE: package = mkcl_core.clos_package; break;
  case GRAY_PACKAGE: package = mkcl_core.gray_package; break;
  default: mkcl_lose("Unknown package in make_this_symbol");
  }
  s->symbol.t = mkcl_t_symbol;
  s->symbol.dynamic = 0;
  s->symbol.special_index = MKCL_NOT_A_SPECIAL_INDEX;
  MKCL_SET(s, MKCL_OBJNULL);
  MKCL_SYM_FUN(s) = mk_cl_Cnil;
  s->symbol.plist = mk_cl_Cnil;
  s->symbol.sys_plist = mk_cl_Cnil;
  s->symbol.hpack = mk_cl_Cnil;
  s->symbol.properly_named_class = mk_cl_Cnil;
  s->symbol.stype = stp;
  s->symbol.hpack = package;
  s->symbol.name = mkcl_make_simple_base_string(env, (char *) name);
  if (package == mkcl_core.keyword_package) {
    mkcl_sethash(env, s->symbol.name, package->pack.external, s);
    MKCL_SET(s, s);
  } else {
    int intern_flag;
    MKCL_SET(s, value);
    if (mkcl_find_symbol(env, s->symbol.name, package, &intern_flag) != mk_cl_Cnil
	&& intern_flag == MKCL_SYMBOL_IS_INHERITED) {
      mkcl_shadowing_import(env, s, package);
    } else {
      mkcl_import2(env, s, package);
    }
    mkcl_export2(env, s, package);
  }
  if (form) {
    s->symbol.stype |= mkcl_stp_special_form;
  } else if (fun) {
    mkcl_object f;
    if (narg >= 0) {
      f = mkcl_make_cfun(env, (mkcl_objectfn_fixed) fun, s, MKCL_OBJNULL, narg, NULL);
    } else {
      f = mkcl_make_cfun_va(env, fun, s, MKCL_OBJNULL, NULL);
    }
    MKCL_SYM_FUN(s) = f;
  }
  mkcl_root_symbols_count = i + 1;
}

void
mkcl_init_all_symbols(MKCL)
{
  int i, code, narg;
  const char *name;
  mkcl_object s, value;
  mkcl_objectfn fun;

  /* We skip NIL and T, thus we start at 2. */
  for (i = 2; mkcl_root_symbols[i].init.name != NULL; i++) {
    s = (mkcl_object)(mkcl_root_symbols + i);
    code = mkcl_root_symbols[i].init.type;
    name = mkcl_root_symbols[i].init.name;
    fun = (mkcl_objectfn)mkcl_root_symbols[i].init.fun;
    narg = mkcl_root_symbols[i].init.narg;
    value = mkcl_root_symbols[i].init.value;
    make_this_symbol(env, i, s, code, name, fun, narg, value);
  }
}

