;;;;  -*- Mode: Lisp; Syntax: Common-Lisp; Package: SYSTEM -*-
;;;;
;;;;  Copyright (c) 2010, Jean-Claude Beaudoin.
;;;;  Copyright by a number of previous anonymous authors
;;;;            presumed to be the same as for the rest of MKCL.
;;;;
;;;;    This program is free software; you can redistribute it and/or
;;;;    modify it under the terms of the GNU Lesser General Public
;;;;    License as published by the Free Software Foundation; either
;;;;    version 3 of the License, or (at your option) any later version.
;;;;
;;;;    See file '../Copyright' for full details.

(defparameter *compile-extra-options* nil)
;;;
;;; If you want to debug the CMP compiler or the lisp part of the runtime
;;; then uncomment this.
;;;
#+(or)
(progn
  (setq c::*delete-files* nil)
  (setq *compile-extra-options* '(:c-file t :data-file t :h-file t))
  (proclaim '(optimize (debug 3))) ;; full debug info
  )
;;(proclaim '(optimize (debug 1))) ;; faster, no debug info.
;;(setq *compile-verbose* t)

;;;
;;; * Add include path to not yet installed headers.

(setq c::*mkcl-include-directory* (truename (pathname ".")) ;; truename is needed by MS-Windows
      c::*mkcl-library-directory* (truename (pathname "."))
      )

;;; -H traces include files in gcc.
;;(setq c::*cc-flags* (concatenate 'base-string "-H " c::*cc-flags*))
;;(setq c::*trace-cc* t)

(defun object-file-pathname (destdir source)
  (make-pathname :host (pathname-host destdir)
		 :device (pathname-device destdir)
		 :directory (pathname-directory destdir)
		 :version nil
		 :defaults (compile-file-pathname source :type :object)))

(defun clean-up (destdir sources)
  (dolist (source sources)
    (let ((object (object-file-pathname destdir source)))
      (when (probe-file object)
	(handler-bind ((condition #'identity))
	  (format t "~&Removing: ~S~%" object) (finish-output)
	  (delete-file object))))
    )
  )



;;;
;;; * Timed compilation facility.
;;;
(defun compile-if-old (destdir sources &rest options)
  (unless (probe-file destdir)
    (si::mkdir destdir #o0777))
  (mapcar #'(lambda (source)
	      (setq source (translate-logical-pathname source))
	      (let ((object (object-file-pathname destdir source)))
		(unless (and (probe-file object)
			     (>= (file-write-date object) (file-write-date source))
			     (>= (file-write-date object) 
				 (file-write-date "./mkcl/mkcl-cmp.h")))
		  (format t "~&(compile-file ~S :output-file ~S~{ ~S~})~%"
			  source object (append options *compile-extra-options*))
		  (unless (apply #'compile-file source
				 :output-file object
				 (append options *compile-extra-options*))
		    (clean-up destdir sources)
		    (si::raw-exit 1)) ;; exit if compilation failed
		  )
		object))
	  sources))


(defun build-substitute-asd-file (name system-attribs #|depends-on|#)
  (with-open-file (*standard-output* (make-pathname :name name :type "asd")
                   :direction :output :if-exists :supersede :if-does-not-exist :create)
    (pprint `(defsystem ,name
               :components ((:bundle ,(string name)))
	       ;; ,@(when depends-on
	       ;; 	   `(:depends-on ,depends-on)
	       ;; 	  )
	       ,@system-attribs
	       ))
    (terpri)))

(defun build-module (name sources &key
			  (builtin nil) ;; deprecated! JCB
			  (dir "")
			  ((:prefix si::*init-function-prefix*) "EXT"))
  (handler-bind 
     ((condition #'(lambda (c)
		     (unless (subtypep (type-of c) 'warning)
		       (format t "~&build-module failed on condition: ~A~%" c)
		       (finish-output)
		       (clean-up dir sources)
		       (si:raw-exit 1)))))
   (let* ((name (string-downcase name)))
     (unless (equalp name "asdf")
       (build-substitute-asd-file name nil))
     (if builtin
	 (let* ((objects (compile-if-old dir sources :fasl-p nil)))
	   (unless (c::build-static-library name :lisp-files objects)
	     (clean-up dir sources)
	     (si::raw-exit 1))
	   )
       (let* ((objects (compile-if-old dir sources :fasl-p nil))
	      ;;(c::*suppress-compiler-messages* nil)
	      ;;(c::*suppress-compiler-warnings* nil)
	      ;;(c::*suppress-compiler-notes* nil)
	      ;;(*compile-verbose* t)
	      ;;(*compile-print* t)
	      )
	 (let (result)
	   (format t "~&(c::build-bundle ~S :lisp-files ~S)" name objects)
	   (setq result (c::build-bundle name :lisp-files objects))
	   (unless result
	     (clean-up dir sources) (si::raw-exit 1)) ;; exit if fasl build failed.
	   )
	 #+unix
	 (progn
	   (format t "~&(c::build-static-library ~S :lisp-files ~S)" name objects)
	   (unless (c::build-static-library name :lisp-files objects)
	     (clean-up dir sources)
	     (si::raw-exit 1)))
	 (terpri)
	 )))))


