/* -*- mode: c -*- */
/*
    mkcl.h -- Main headers for use and/or development of MKCL
*/
/*
    Copyright (c) 1984, Taiichi Yuasa and Masami Hagiya.
    Copyright (c) 1990, Giuseppe Attardi.
    Copyright (c) 2010-2011, Jean-Claude Beaudoin.

    MKCL is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    See file './Copyright' for full details.
*/

#ifndef MKCL_H
#define MKCL_H

#ifdef __MINGW32__
#define mingw32
#endif

#ifdef mingw32
/* This define has to be done before any system include. */
# define __USE_MINGW_ANSI_STDIO 1
#endif

#include <sys/types.h>

#include <signal.h>
#include <stddef.h>		/* NULL, ptrdiff_t */
#include <stdarg.h> 		/* va_list */
#include <setjmp.h> 		/* setjmp and buffers */
#include <stdio.h>		/* FILE */


/* Microsoft VC++ does not have va_copy() */
#if defined(_MSC_VER) || !defined(va_copy)
# define va_copy(dst, src) \
   ((void) memcpy(&(dst), &(src), sizeof(va_list)))
#endif

#include <mkcl/config.h>

#ifdef MKCL_WINDOWS
# define WIN32_LEAN_AND_MEAN 1 /* Do not include winsock.h */
# include <windows.h>
typedef HANDLE pthread_t;
typedef HANDLE pthread_mutex_t;
# ifdef __MINGW64__
typedef int sigset_t; /* This is obviously a non-functional stub. */
# endif
# if defined(_MSC_VER) && (defined(WinVista) || defined(Win7))
typedef SRWLOCK pthread_rwlock_t; /* MingW does not support this yet it seems. JCB 2010/02/15 */
# else
typedef HANDLE pthread_rwlock_t;
# endif
typedef HANDLE sem_t;
typedef HANDLE pthread_cond_t; /* Dummy, not really used */
#else
# include <pthread.h>
# include <semaphore.h>
#endif

#ifndef __GNUC__
# define mkcl_likely(expr) (expr)
# define mkcl_unlikely(expr) (expr)
# define mkcl_attr_noreturn
#else
# define mkcl_likely(expr) __builtin_expect((expr), TRUE)
# define mkcl_unlikely(expr) __builtin_expect((expr), FALSE)
# define mkcl_noreturn __attribute__((noreturn))
# if (__GNUC__ < 4)
#  error MKCL expects GCC version 4.X.X or later
# endif
#endif


#include <mkcl/object.h>
#include <mkcl/number.h>
#include <mkcl/list.h>
#include <mkcl/external.h>
#include <mkcl/stacks.h>

#endif /* MKCL_H */
