;; benchmarks speed of array and sequence operations
;;
;; Author: Jean-Claude Beaudoin
;; Time-stamp: <2011-03-22 jean-claude>
;;
;; Derived from 
;; Author: Eric Marsden <emarsden@laas.fr>
;; Time-stamp: <2003-12-29 emarsden>
;;
;;
;; Timing tests for creation, initialization, access and garbage
;; collection for arrays, vectors, bitvectors and strings.
;;

;;(in-package :cl-bench.arrays)

(deftype index ()
  `(INTEGER 0 (#.array-dimension-limit)))


;;(declaim (optimize (speed 3) (space 1) (safety 0) (debug 0) (compilation-speed 0)))
(declaim (optimize (speed 3) (space 1) (debug 0)))


(defun bench-1d-t-arrays (&optional (size 100000) (runs 100))
  (declare (fixnum size runs))
  (let ((ones (make-array size :initial-element 1))
        (twos (make-array size :initial-element 2))
        (threes (make-array size)))
    (declare (simple-vector ones twos threes))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (pos size)
        (declare (index pos))
        (setf (aref threes pos) (the fixnum (+ (the fixnum (aref ones pos)) (the fixnum (aref twos pos))))))
      ;;(assert (null (search (list 4 5 6) threes)))
      ))
  (values))

(defun bench-1d-arrays (&optional (size 100000) (runs 100))
  (declare (fixnum size runs))
  (let ((ones (make-array size :element-type 'fixnum :initial-element 1))
        (twos (make-array size :element-type 'fixnum :initial-element 2))
        (threes (make-array size :element-type 'fixnum)))
    (declare (type (vector fixnum) ones twos) (type (vector fixnum) threes))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (pos size)
	(declare (index pos))
        (setf (aref threes pos) (the fixnum (+ (aref ones pos) (aref twos pos)))))
      ;;(assert (null (search (list 4 5 6) threes)))
      ))
  (values))

(defun bench-1d-arrays+assert (&optional (size 100000) (runs 100))
  (declare (fixnum size runs))
  (let ((ones (make-array size :element-type 'fixnum :initial-element 1))
        (twos (make-array size :element-type 'fixnum :initial-element 2))
        (threes (make-array size :element-type 'fixnum)))
    (declare (type (vector fixnum) ones twos) (type (vector fixnum) threes))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (pos size)
	(declare (index pos))
        (setf (aref threes pos) (the fixnum (+ (aref ones pos) (aref twos pos)))))
      (assert (null (search (list 4 5 6) threes)))))
  (values))

(defun bench-2d-arrays (&optional (size 2000) (runs 10))
  (declare (fixnum size runs))
  (let ((ones (make-array (list size size) :element-type 'fixnum :initial-element 1))
        (twos (make-array (list size size) :element-type 'fixnum :initial-element 2))
        (threes (make-array (list size size) :element-type 'fixnum)))
    (declare (type (array fixnum (* *)) ones twos) (type (array fixnum (* *)) threes))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (i size)
	(declare (index i))
        (dotimes (j size)
	  (declare (index j))
	  #+(or)
	  (setf (row-major-aref threes (array-row-major-index threes i j))
		(the fixnum (+ (row-major-aref ones (array-row-major-index ones i j))
					 (row-major-aref twos (array-row-major-index twos i j)))))
	  #+(and)
          (setf (aref threes i j)
                (the fixnum (+ (aref ones i j) (aref twos i j))))
	  ))
      ;;(assert (eql 3 (aref threes 3 3)))
      ))
  (values))

(defun bench-2d-arrays+assert (&optional (size 2000) (runs 10))
  (declare (fixnum size runs))
  (let ((ones (make-array (list size size) :element-type 'fixnum :initial-element 1))
        (twos (make-array (list size size) :element-type 'fixnum :initial-element 2))
        (threes (make-array (list size size) :element-type 'fixnum)))
    (declare (type (array fixnum (* *)) ones twos) (type (array fixnum (* *)) threes))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (i size)
	(declare (index i))
        (dotimes (j size)
	  (declare (index j))
	  #+(and)
	  (setf (row-major-aref threes (array-row-major-index threes i j))
		(the fixnum (+ (row-major-aref ones (array-row-major-index ones i j))
					 (row-major-aref twos (array-row-major-index twos i j)))))
	  #+(or)
          (setf (aref threes i j)
                (the fixnum (+ (the fixnum (aref ones i j)) (the fixnum (aref twos i j)))))))
      (assert (eql 3 (aref threes 3 3)))))
  (values))

(defun bench-3d-arrays (&optional (size 200) (runs 10))
  (declare (fixnum size runs))
  (let ((ones (make-array (list size size size) :element-type 'fixnum :initial-element 1))
        (twos (make-array (list size size size) :element-type 'fixnum :initial-element 2))
        (threes (make-array (list size size size) :element-type 'fixnum)))
    (declare (type (array fixnum (* * *)) ones twos) (type (array fixnum (* * *)) threes))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (i size)
	(declare (index i))
        (dotimes (j size)
	  (declare (index j))
          (dotimes (k size)
	    (declare (index k))
	    #+(and)
	    (setf (row-major-aref threes (array-row-major-index threes i j k))
		  (the fixnum (+ (row-major-aref ones (array-row-major-index ones i j k))
					   (row-major-aref twos (array-row-major-index twos i j k)))))
	    #+(or)
            (setf (aref threes i j k)
                  (the fixnum (+ (the fixnum (aref ones i j k)) (the fixnum (aref twos i j k)))))
	    )))
      ;(assert (eql 3 (aref threes 3 3 3)))
      ))
  (values))

(defun bench-bitvectors (&optional (size 1000000) (runs 700))
  (declare (fixnum size runs))
  (let ((zeros (make-array size :element-type 'bit :initial-element 0))
        (ones  (make-array size :element-type 'bit :initial-element 1))
        (xors  (make-array size :element-type 'bit)))
    (declare (type (vector bit) zeros ones xors))
    (dotimes (run runs)
      (declare (index run))
      (bit-xor zeros ones xors)
      (bit-nand zeros ones xors)
      (bit-and zeros xors)))
  (values))

(defun bench-strings (&optional (size 1000000) (runs 5000))
  (declare (fixnum size runs))
  (let ((zzz (make-string size :initial-element #\z))
        (xxx (make-string size)))
    (declare (type simple-string zzz xxx))
    (dotimes (run runs)
      (declare (index run))
      (and (fill xxx #\x)
           (replace xxx zzz)
           (search "xxxd" xxx)
           (progn (format t "~&about to call nstring-upcase!~%")(nstring-upcase xxx)))))
  (values))

(defun bench-strings/adjustable (&optional (size 1000000) (runs 100) &aux next)
  (declare (fixnum size runs))
  (dotimes (run runs)
    (declare (index run))
    (let ((sink (make-array 10 :element-type 'character :adjustable t :fill-pointer 0)))
      (declare (type string sink))
      (dotimes (i size)
	(declare (index i))
        (setq next (vector-push-extend (code-char (mod i 128)) sink)))))
  (values next))

;; certain implementations such as OpenMCL have an array (and thus
;; string) length limit of (expt 2 24), so don't try this on humungous
;; sizes
(defun bench-string-concat (&optional (size 1000000) (runs 100))
  (declare (fixnum size runs))
  (dotimes (run runs)
    (declare (index run))
    (let ((len (length
                (with-output-to-string (string)
                  (dotimes (i size)
		    (declare (index i))
                    (write-sequence "hi there!" string))))))
      len
      (assert (eql len (* size (length "hi there!"))))
      )
    (values)))

(defun bench-search-sequence (&optional (size 1000000) (runs 10))
  (declare (fixnum size runs))
  (let ((haystack (make-array size :element-type 'fixnum)))
    (declare (type (vector fixnum) haystack))
    (dotimes (run runs)
      (declare (index run))
      (dotimes (i size)
	(declare (index i))
        (setf (aref haystack i) (mod i 1000)))
      (assert (null (find -1 haystack :test #'=)))
      (assert (null (find-if #'minusp haystack)))
      (assert (null (position -1 haystack :test #'= :from-end t)))
      (loop :for i :from 20 :to 900 :by 20
            :do (assert (eql i (position i haystack :test #'=))))
      (assert (eql 0 (search #(0 1 2 3 4) haystack :end2 1000 :from-end t)))))
  (values))


;; EOF

(defun full-gc ()
  #+mkcl (si:gc)
  #+ecl (si:gc t)
  #+sbcl (sb-ext:gc #+gencgc :full #+gencgc t)
  #+clisp (gc)
  )
  

(defun run-test ()
  (format t "~&Running bench-1d-t-arrays.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-1d-t-arrays)))
  (format t "~&Running bench-1d-arrays.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-1d-arrays)))
  (format t "~&Running bench-1d-arrays+assert.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-1d-arrays+assert)))
  (format t "~&Running bench-2d-arrays.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-2d-arrays)))
  (format t "~&Running bench-2d-arrays+assert.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-2d-arrays+assert)))
  (format t "~&Running bench-3d-arrays.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-3d-arrays)))
  (format t "~&Running bench-bitvectors.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-bitvectors)))
  (format t "~&Running bench-strings.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-strings)))
  (format t "~&Running bench-strings/adjustable.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-strings/adjustable)))
  (format t "~&Running bench-string-concat.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-string-concat)))
  (format t "~&Running bench-search-sequence.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 1) (bench-search-sequence)))
)

