
;(declaim (optimize (debug 0) (safety 0)))
(declaim (optimize (debug 0)))

(defun fix-list (fix)
  (declare (fixnum fix))
  (list fix))


;;; CTAK -- A version of the TAKeuchi function that uses the CATCH/THROW facility.

(defun ctak-aux (x y z)
  (declare (fixnum x y z))
  (cond ((not (< y x))
	 (throw 'ctak z))
	(t (ctak-aux
	     (catch 'ctak
	       (ctak-aux (the fixnum (1- x))
			 y
			 z))
	     (catch 'ctak
	       (ctak-aux (the fixnum (1- y))
			 z
			 x))
	     (catch 'ctak
	       (ctak-aux (the fixnum (1- z))
			 x
			 y))))))

(defun ctak (x y z)
  (declare (fixnum x y z))
  (catch 'ctak (ctak-aux x y z)))

(defun run-ctak ()
  (ctak 22 12 8))

;;; TAK -- A vanilla version of the TAKeuchi function and one with tail recursion
;;; removed.

;#-mkcl
(defun tak (x y z)
  (declare (fixnum x y z))
  (if (not (< y x))
      z
      (tak (tak (the fixnum (1- x)) y z)
	   (tak (the fixnum (1- y)) z x)
	   (tak (the fixnum (1- z)) x y))))

#+(or) ;;mkcl
(defun tak (x y z)
  (if (not (< (the fixnum y) (the fixnum x)))
      z
      (tak (tak ;(the fixnum (1- (the fixnum x)))
		(1- (the fixnum x))
		y z)
	   (tak ;(the fixnum (1- (the fixnum y)))
		(1- (the fixnum y))
		z x)
	   (tak ;(the fixnum (1- (the fixnum z)))
		(1- (the fixnum z))
		x y))))

(defun run-tak ()
  (tak 18 12 6))


(defun run-ltak ()
  (labels ((ltak (x y z)
	     (declare (fixnum x y z))
	     (if (not (< y x))
		 z
	       (the fixnum (ltak (the fixnum (ltak (the fixnum (1- x)) y z))
				 (the fixnum (ltak (the fixnum (1- y)) z x))
				 (the fixnum (ltak (the fixnum (1- z)) x y)))))))
   (declare (ftype (function (fixnum fixnum fixnum) fixnum) ltak))
   (ltak 18 12 6))
  )


(defun trtak (x y z)
  (declare (fixnum x y z))
  (prog ()
    tak
    (if (not (< y x))
	(return z)
	(let ((a (trtak (the fixnum (1- x)) y z))
	      (b (trtak (the fixnum (1- y)) z x)))
	  (declare (fixnum a b))
	  (setq z (the fixnum (trtak (1- z) x y))
		x a
		y b)
	  (go tak)))))

(defun run-trtak ()
   (trtak 18 12 6))


;;; TAKL -- The TAKeuchi function using lists as counters.

(defun listn (n)
  (if (not (= 0 (the fixnum n)))
      (cons n (listn (the fixnum (1- (the fixnum n)))))))

(defun shorterp (x y)
  (declare (list x y))
  (and y (or (null x)
	     (shorterp (cdr x)
		       (cdr y)))))

(defun mas (x y z)
  (declare (list x y z))
  (if (not (shorterp y x))
      z
      (mas (mas (cdr x)
		 y z)
	    (mas (cdr y)
		 z x)
	    (mas (cdr z)
		 x y))))

(defun run-takl ()
  (let ((l18 (listn 18.))
        (l12 (listn 12.))
        (l6 (listn 6.)))
    (mas l18 l12 l6)))


;;; STAK -- The TAKeuchi function with special variables instead of parameter
;;; passing.

(defvar *stak-x*)
(defvar *stak-y*)
(defvar *stak-z*)

(defun stak-aux ()
  (if (not (< (the fixnum *stak-y*) (the fixnum *stak-x*)))
      *stak-z*
      (let ((*stak-x* (let ((*stak-x* (the fixnum (- (the fixnum *stak-x*) 1)))
                            (*stak-y* *stak-y*)
                            (*stak-z* *stak-z*))
                        (stak-aux)))
	    (*stak-y* (let ((*stak-x* (the fixnum (- (the fixnum *stak-y*) 1)))
                            (*stak-y* *stak-z*)
                            (*stak-z* *stak-x*))
                        (stak-aux)))
	    (*stak-z* (let ((*stak-x* (the fixnum (- (the fixnum *stak-z*) 1)))
                            (*stak-y* *stak-x*)
                            (*stak-z* *stak-y*))
                        (stak-aux))))
	(stak-aux))))

(defun stak (*stak-x* *stak-y* *stak-z*)
  (stak-aux))

(defun run-stak ()
  (stak 18 12 6))



(defun full-gc ()
  #+mkcl (si:gc)
  #+ecl (si:gc t)
  #+sbcl (sb-ext:gc #+gencgc :full #+gencgc t)
  #+clisp (gc)
  )

(defun run-test ()
  (format t "~&Running run-tak.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 500) (declare (fixnum i)) (run-tak)))
  (format t "~&Running run-ctak.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 500) (declare (fixnum i)) (run-ctak)))
  (format t "~&Running run-ltak.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 500) (declare (fixnum i)) (run-ltak)))
  (format t "~&Running run-takl.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 500) (declare (fixnum i)) (run-takl)))
  (format t "~&Running run-stak.~%") (finish-output)
  (full-gc)
  (time (dotimes (i 500) (declare (fixnum i)) (run-stak)))
  )
